"""
    Copyright 2017 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com


    Command line development guidelines
    ###################################

    do's and don'ts
    ----------------
    MUST NOT: sys.exit => use command.CLIException
    SHOULD NOT: print( => use logger for messages, only print for final output


    Entry points
    ------------
    @command annotation to register new command
"""

from argparse import ArgumentParser
import logging
import sys
import time
import json
import os
import socket
import signal

import colorlog

from inmanta.command import command, Commander, CLIException
from inmanta.compiler import do_compile
from inmanta.config import Config
from tornado.ioloop import IOLoop
from tornado.util import TimeoutError
from tornado import gen
from inmanta import protocol, module, moduletool, const
from inmanta.export import cfg_env, ModelExporter
import yaml

from inmanta.server.bootloader import InmantaBootloader
from inmanta.ast import CompilerException
import asyncio
import traceback
import threading
from threading import Timer

LOGGER = logging.getLogger()


@command("server", help_msg="Start the inmanta server")
def start_server(options):
    ibl = InmantaBootloader()
    setup_signal_handlers(ibl.stop)
    IOLoop.current().add_callback(ibl.start)
    IOLoop.current().start()
    print("Server Shutdown complete")


@command("agent", help_msg="Start the inmanta agent")
def start_agent(options):
    from inmanta import agent
    a = agent.Agent()
    setup_signal_handlers(a.stop)
    IOLoop.current().add_callback(a.start)
    IOLoop.current().start()
    print("Agent Shutdown complete")


def dump_threads():
    print("----- Thread Dump ----")
    for th in threading.enumerate():
        print("---", th)
        traceback.print_stack(sys._current_frames()[th.ident])
        print()
    sys.stdout.flush()


@gen.coroutine
def dump_ioloop_running():
    # dump async IO
    print("----- Async IO tasks ----")
    for task in asyncio.all_tasks():
        print(task)
    print()
    sys.stdout.flush()


def context_dump(ioloop):
    dump_threads()
    if hasattr(asyncio, "all_tasks"):
        ioloop.add_callback_from_signal(dump_ioloop_running)


def setup_signal_handlers(shutdown_function):
    """
        Make sure that shutdown_function is called when a SIGTERM or a SIGINT interrupt occurs.

        :param shutdown_function: The function that contains the shutdown logic.
    """
    # ensure correct ioloop
    ioloop = IOLoop.current()

    def hard_exit():
        context_dump(ioloop)
        sys.stdout.flush()
        # Hard exit, not sys.exit
        # ensure shutdown when the ioloop is stuck
        os._exit(const.EXIT_HARD)

    def handle_signal(signum, frame):
        # force shutdown, even when the ioloop is stuck
        # schedule off the loop
        t = Timer(const.SHUTDOWN_GRACE_HARD, hard_exit)
        t.daemon = True
        t.start()
        ioloop.add_callback_from_signal(safe_shutdown_wrapper, shutdown_function)

    def handle_signal_dump(signum, frame):
        context_dump(ioloop)

    signal.signal(signal.SIGTERM, handle_signal)
    signal.signal(signal.SIGINT, handle_signal)
    signal.signal(signal.SIGUSR1, handle_signal_dump)


@gen.coroutine
def safe_shutdown_wrapper(shutdown_function):
    """
        Wait 10 seconds to gracefully shutdown the instance.
        Afterwards stop the IOLoop
        Wait for 3 seconds to force stop
    """
    future = shutdown_function()
    try:
        timeout = IOLoop.current().time() + const.SHUTDOWN_GRACE_IOLOOP
        yield gen.with_timeout(timeout, future)
    except TimeoutError:
        pass
    finally:
        IOLoop.current().stop()


def compiler_config(parser):
    """
        Configure the compiler of the export function
    """
    parser.add_argument("-e", dest="environment", help="The environment to compile this model for")
    parser.add_argument("-X", "--extended-errors", dest="errors",
                        help="Show stack traces for compile errors", action="store_true", default=False)
    parser.add_argument("--server_address", dest="server", help="The address of the server hosting the environment")
    parser.add_argument("--server_port", dest="port", help="The port of the server hosting the environment")
    parser.add_argument("--username", dest="user", help="The username of the server")
    parser.add_argument("--password", dest="password", help="The password of the server")
    parser.add_argument("--ssl", help="Enable SSL", action="store_true", default=False)
    parser.add_argument("--ssl-ca-cert", dest="ca_cert", help="Certificate authority for SSL")
    parser.add_argument("-f", dest="main_file", help="Main file", default="main.cf")


@command("compile", help_msg="Compile the project to a configuration model",
         parser_config=compiler_config, require_project=True)
def compile_project(options):
    if options.environment is not None:
        Config.set("config", "environment", options.environment)

    if options.server is not None:
        Config.set("compiler_rest_transport", "host", options.server)

    if options.port is not None:
        Config.set("compiler_rest_transport", "port", options.port)

    if options.user is not None:
        Config.set("compiler_rest_transport", "username", options.user)

    if options.password is not None:
        Config.set("compiler_rest_transport", "password", options.password)

    if options.ssl:
        Config.set("compiler_rest_transport", "ssl", "true")

    if options.ca_cert is not None:
        Config.set("compiler_rest_transport", "ssl-ca-cert-file", options.ca_cert)

    module.Project.get(options.main_file)

    if options.profile:
        import cProfile
        import pstats
        result = cProfile.runctx('do_compile()', globals(), {}, "run.profile")
        p = pstats.Stats('run.profile')
        p.strip_dirs().sort_stats("time").print_stats(20)
    else:
        t1 = time.time()
        result = do_compile()
        LOGGER.debug("Compile time: %0.03f seconds", time.time() - t1)
    return result


@command("list-commands", help_msg="Print out an overview of all commands")
def list_commands(options):
    print("The following commands are available:")
    for cmd, info in Commander.commands().items():
        print(" %s: %s" % (cmd, info["help"]))


def help_parser_config(parser: ArgumentParser):
    parser.add_argument("subcommand", help="Output help for a particular subcommand",
                        nargs="?", default=None)


@command("help", help_msg="show a help message and exit", parser_config=help_parser_config)
def help_command(options):
    if options.subcommand is None:
        cmd_parser().print_help()
    else:
        subc = options.subcommand
        parser = cmd_parser()
        parser.parse_args([subc, "-h"])
    sys.exit(0)


@command("modules", help_msg="Subcommand to manage modules",
         parser_config=moduletool.ModuleTool.modules_parser_config,
         aliases=["module"])
def modules(options):
    tool = moduletool.ModuleTool()
    tool.execute(options.cmd, options)


@command("project", help_msg="Subcommand to manage the project",
         parser_config=moduletool.ProjectTool.parser_config)
def project(options):
    tool = moduletool.ProjectTool()
    tool.execute(options.cmd, options)


def deploy_parser_config(parser):
    parser.add_argument("--dry-run", help="Only report changes", action="store_true", dest="dryrun")
    parser.add_argument("-f", dest="main_file", help="Main file", default="main.cf")
    parser.add_argument(
        "--dashboard",
        dest="dashboard",
        help="Start the dashboard and keep the server running. "
             "The server uses the current project as the source for server recompiles",
        action="store_true",
        default=False
    )


@command("deploy", help_msg="Deploy with a inmanta all-in-one setup", parser_config=deploy_parser_config, require_project=True)
def deploy(options):
    module.Project.get(options.main_file)
    from inmanta import deploy

    run = deploy.Deploy(options)
    try:
        run.setup()
        run.run()
    finally:
        run.stop()


def export_parser_config(parser):
    """
        Configure the compiler of the export function
    """
    parser.add_argument("-g", dest="depgraph", help="Dump the dependency graph", action="store_true")
    parser.add_argument("-j", dest="json", help="Do not submit to the server but only store the json that would have been "
                                                "submitted in the supplied file")
    parser.add_argument("-e", dest="environment", help="The environment to compile this model for")
    parser.add_argument("-d", dest="deploy", help="Trigger a deploy for the exported version", action="store_true")
    parser.add_argument("--full", dest="full_deploy",
                        help="Make the agents execute a full deploy instead of an incremental deploy. "
                             "Should be used together with the -d option",
                        action="store_true", default=False)
    parser.add_argument("-m", dest="model", help="Also export the complete model", action="store_true", default=False)
    parser.add_argument("--server_address", dest="server", help="The address of the server to submit the model to")
    parser.add_argument("--server_port", dest="port", help="The port of the server to submit the model to")
    parser.add_argument("--token", dest="token", help="The token to auth to the server")
    parser.add_argument("--ssl", help="Enable SSL", action="store_true", default=False)
    parser.add_argument("--ssl-ca-cert", dest="ca_cert", help="Certificate authority for SSL")
    parser.add_argument("-X", "--extended-errors", dest="errors",
                        help="Show stack traces for compile errors", action="store_true", default=False)
    parser.add_argument("-f", dest="main_file", help="Main file", default="main.cf")
    parser.add_argument("--metadata", dest="metadata", help="JSON metadata why this compile happened. If a non-json string is "
                        "passed it is used as the 'message' attribute in the metadata.",
                        default=None)
    parser.add_argument("--model-export", dest="model_export", help="Export the configuration model to the server as metadata.",
                        action="store_true", default=False)


@command("export", help_msg="Export the configuration", parser_config=export_parser_config, require_project=True)
def export(options):
    if options.environment is not None:
        Config.set("config", "environment", options.environment)

    if options.server is not None:
        Config.set("compiler_rest_transport", "host", options.server)

    if options.server is not None:
        Config.set("compiler_rest_transport", "port", options.port)

    if options.token is not None:
        Config.set("compiler_rest_transport", "token", options.token)

    if options.ssl:
        Config.set("compiler_rest_transport", "ssl", "true")

    if options.ca_cert is not None:
        Config.set("compiler_rest_transport", "ssl-ca-cert-file", options.ca_cert)

    # try to parse the metadata as json. If a normal string, create json for it.
    if options.metadata is not None and len(options.metadata) > 0:
        try:
            metadata = json.loads(options.metadata)
        except json.decoder.JSONDecodeError:
            metadata = {"message": options.metadata}
    else:
        metadata = {"message": "Manual compile on the CLI by user"}

    if "cli-user" not in metadata and "USERNAME" in os.environ:
        metadata["cli-user"] = os.environ["USERNAME"]

    if "hostname" not in metadata:
        metadata["hostname"] = socket.gethostname()

    if "type" not in metadata:
        metadata["type"] = "manual"

    module.Project.get(options.main_file)

    from inmanta.export import Exporter  # noqa: H307

    exp = None
    try:
        (types, scopes) = do_compile()
    except Exception as e:
        exp = e
        types, scopes = (None, None)

    # Even if the compile failed we might have collected additional data such as unknowns. So
    # continue the export

    export = Exporter(options)
    version, _ = export.run(types, scopes, metadata=metadata, model_export=options.model_export)

    if exp is not None:
        raise exp

    if options.model:
        modelexporter = ModelExporter(types)
        with open("testdump.json", "w") as fh:
            print(yaml.dump(modelexporter.export_all()))
            json.dump(modelexporter.export_all(), fh)

    if options.deploy:
        conn = protocol.SyncClient("compiler")
        LOGGER.info("Triggering deploy for version %d" % version)
        tid = cfg_env.get()
        agent_trigger_method = const.AgentTriggerMethod.get_agent_trigger_method(options.full_deploy)
        conn.release_version(tid, version, True, agent_trigger_method)


log_levels = {
    0: logging.ERROR,
    1: logging.WARNING,
    2: logging.INFO,
    3: logging.DEBUG,
    4: 2
}


def cmd_parser():
    # create the argument compiler
    parser = ArgumentParser()
    parser.add_argument("-p", action="store_true", dest="profile", help='Profile this run of the program')
    parser.add_argument("-c", "--config", dest="config_file", help="Use this config file")
    parser.add_argument("--log-file", dest="log_file", help="Path to the logfile")
    parser.add_argument("--log-file-level", dest="log_file_level", default=2, type=int,
                        help="Log level for messages going to the logfile: 0=ERROR, 1=WARNING, 2=INFO, 3=DEBUG")
    parser.add_argument("--timed-logs", dest="timed", help="Add timestamps to logs", action="store_true")
    parser.add_argument('-v', '--verbose', action='count', default=0,
                        help="Log level for messages going to the console. Default is only errors,"
                        "-v warning, -vv info and -vvv debug and -vvvv trace")
    parser.add_argument("-X", "--extended-errors", dest="errors",
                        help="Show stack traces for errors", action="store_true", default=False)
    subparsers = parser.add_subparsers(title="commands")
    for cmd_name, cmd_options in Commander.commands().items():
        cmd_subparser = subparsers.add_parser(cmd_name, help=cmd_options["help"], aliases=cmd_options["aliases"])
        if cmd_options["parser_config"] is not None:
            cmd_options["parser_config"](cmd_subparser)
        cmd_subparser.set_defaults(func=cmd_options["function"])
        cmd_subparser.set_defaults(require_project=cmd_options["require_project"])

    return parser


def _is_on_tty() -> bool:
    return (hasattr(sys.stdout, 'isatty') and sys.stdout.isatty()) or const.ENVIRON_FORCE_TTY in os.environ


def _get_default_stream_handler():
    stream_handler = logging.StreamHandler(stream=sys.stdout)
    stream_handler.setLevel(logging.INFO)

    formatter = _get_log_formatter_for_stream_handler(timed=False)
    stream_handler.setFormatter(formatter)

    return stream_handler


def _get_watched_file_handler(options):
    if not options.log_file:
        raise Exception("No logfile was provided.")
    level = _convert_to_log_level(options.log_file_level)
    formatter = logging.Formatter(fmt="%(asctime)s %(levelname)-8s %(name)-10s %(message)s")
    file_handler = logging.handlers.WatchedFileHandler(filename=options.log_file, mode='a+')
    file_handler.setFormatter(formatter)
    file_handler.setLevel(level)

    return file_handler


def _convert_to_log_level(level):
    if level >= len(log_levels):
        level = len(log_levels) - 1
    return log_levels[level]


def _get_log_formatter_for_stream_handler(timed):
    log_format = "%(asctime)s " if timed else ""
    if _is_on_tty():
        log_format += "%(log_color)s%(name)-25s%(levelname)-8s%(reset)s %(blue)s%(message)s"
        formatter = colorlog.ColoredFormatter(
            log_format,
            datefmt=None,
            reset=True,
            log_colors={
                'DEBUG': 'cyan',
                'INFO': 'green',
                'WARNING': 'yellow',
                'ERROR': 'red',
                'CRITICAL': 'red',
            }
        )
    else:
        log_format += "%(name)-25s%(levelname)-8s%(message)s"
        formatter = logging.Formatter(fmt=log_format)
    return formatter


def app():
    """
        Run the compiler
    """
    # Send logs to stdout
    stream_handler = _get_default_stream_handler()
    logging.root.handlers = []
    logging.root.addHandler(stream_handler)
    logging.root.setLevel(0)

    # do an initial load of known config files to build the libdir path
    Config.load_config()
    parser = cmd_parser()
    options, other = parser.parse_known_args()
    options.other = other

    # Log everything to a log_file if logfile is provided
    if options.log_file:
        watched_file_handler = _get_watched_file_handler(options)
        logging.root.addHandler(watched_file_handler)
        logging.root.removeHandler(stream_handler)
    else:
        if options.timed:
            formatter = _get_log_formatter_for_stream_handler(timed=True)
            stream_handler.setFormatter(formatter)
        log_level = _convert_to_log_level(options.verbose)
        stream_handler.setLevel(log_level)

    logging.captureWarnings(True)

    # Load the configuration
    Config.load_config(options.config_file)

    # start the command
    if not hasattr(options, "func"):
        # show help
        parser.print_usage()
        return

    def report(e):
        if not options.errors:
            if isinstance(e, CompilerException):
                print(e.format_trace(indent="  "), file=sys.stderr)
            else:
                print(str(e), file=sys.stderr)
        else:
            sys.excepthook(*sys.exc_info())

        if isinstance(e, CompilerException):
            from inmanta.compiler.help.explainer import ExplainerFactory
            helpmsg = ExplainerFactory().explain_and_format(e, plain=not _is_on_tty())
            if helpmsg is not None:
                print(helpmsg)

    try:
        options.func(options)
    except CLIException as e:
        report(e)
        sys.exit(e.exitcode)
    except Exception as e:
        report(e)
        sys.exit(1)
    except KeyboardInterrupt as e:
        report(e)
        sys.exit(1)
    sys.exit(0)


if __name__ == "__main__":
    app()



"""
    Copyright 2016 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""
import uuid

import pytest


@pytest.mark.asyncio
async def test_project_api(client):
    result = await client.create_project("project-test")
    assert result.code == 200
    assert "project" in result.result
    assert "id" in result.result["project"]

    project_id = result.result["project"]["id"]

    result = await client.create_project("project-test")
    assert result.code == 500

    result = await client.list_projects()
    assert result.code == 200
    assert "projects" in result.result
    assert len(result.result["projects"]) == 1

    assert result.result["projects"][0]['id'] == project_id

    result = await client.get_project(id=project_id)
    assert result.code == 200
    assert "project" in result.result
    assert result.result["project"]['id'] == project_id
    assert result.result["project"]['name'] == "project-test"

    result = await client.modify_project(id=project_id, name="project-test2")
    assert result.code == 200
    assert "project" in result.result
    assert result.result["project"]['id'] == project_id
    assert result.result["project"]['name'] == "project-test2"

    result = await client.get_project(id=project_id)
    assert result.code == 200
    assert "project" in result.result
    assert result.result["project"]['id'] == project_id
    assert result.result["project"]['name'] == "project-test2"

    result = await client.delete_project(id=project_id)
    assert result.code == 200

    result = await client.list_projects()
    assert result.code == 200
    assert "projects" in result.result
    assert len(result.result["projects"]) == 0


@pytest.mark.asyncio
async def test_env_api(client):
    result = await client.create_project("env-test")
    assert result.code == 200
    assert "project" in result.result
    assert "id" in result.result["project"]
    project_id = result.result["project"]["id"]

    result = await client.create_environment(project_id=project_id, name="dev")
    assert result.code == 200
    assert "environment" in result.result
    assert "id" in result.result["environment"]
    assert "project" in result.result["environment"]
    assert project_id == result.result["environment"]["project"]
    assert "dev" == result.result["environment"]["name"]

    env_id = result.result["environment"]["id"]

    result = await client.modify_environment(id=env_id, name="dev2")
    assert result.code == 200
    assert "environment" in result.result
    assert result.result["environment"]['id'] == env_id
    assert result.result["environment"]['name'] == "dev2"

    result = await client.get_environment(id=env_id)
    assert result.code == 200
    assert "environment" in result.result
    assert result.result["environment"]['id'] == env_id
    assert result.result["environment"]['project'] == project_id
    assert result.result["environment"]['name'] == "dev2"

    project_result = await client.get_project(id=project_id)
    assert project_result.code == 200
    assert "project" in project_result.result
    assert env_id in project_result.result["project"]["environments"]

    result = await client.list_environments()
    assert result.code == 200
    assert len(result.result) == 1

    result = await client.delete_environment(id=env_id)
    assert result.code == 200

    result = await client.list_environments()
    assert result.code == 200
    assert "environments" in result.result
    assert len(result.result["environments"]) == 0


@pytest.mark.asyncio
async def test_project_cascade(client):
    result = await client.create_project("env-test")
    project_id = result.result["project"]["id"]

    result = await client.create_environment(project_id=project_id, name="dev")
    result = await client.create_environment(project_id=project_id, name="prod")

    result = await client.delete_project(project_id)
    assert result.code == 200

    result = await client.list_environments()
    assert len(result.result["environments"]) == 0


@pytest.mark.asyncio
async def test_create_with_id(client):
    project_id = uuid.uuid4()
    result = await client.create_project(name="test_project", project_id=project_id)
    assert result.result["project"]["id"] == str(project_id)

    env_id = uuid.uuid4()
    result = await client.create_environment(project_id=project_id, name="test_env", environment_id=env_id)
    env = result.result["environment"]
    assert env["id"] == str(env_id)
    assert env["project"] == str(project_id)

    result = await client.create_environment(project_id=project_id, name="test_env2", environment_id=env_id)
    assert result.code == 500



"""
    Copyright 2018 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""
import pytest

from inmanta.ast import AttributeException
from inmanta.ast import RuntimeException, OptionalValueException
import inmanta.compiler as compiler
from inmanta.parser import ParserException


def test_list_atributes(snippetcompiler):
    snippetcompiler.setup_for_snippet(
        """
entity Jos:
  bool[] bar
  std::package_state[] ips = ["installed"]
  string[] floom = []
  string[] floomx = ["a", "b"]
  string box = "a"
end

implement Jos using std::none

a = Jos(bar = [true])
b = Jos(bar = [true, false])
c = Jos(bar = [])
d = Jos(bar = [], floom=["test","test2"])

"""
    )
    (_, root) = compiler.do_compile()

    def check_jos(jos, bar, ips=["installed"], floom=[], floomx=["a", "b"], box="a"):
        jos = jos.get_value()
        assert jos.get_attribute("bar").get_value() == bar
        assert jos.get_attribute("ips").get_value(), ips
        assert jos.get_attribute("floom").get_value() == floom
        assert jos.get_attribute("floomx").get_value() == floomx
        assert jos.get_attribute("box").get_value() == box

    scope = root.get_child("__config__").scope

    check_jos(scope.lookup("a"), [True])
    check_jos(scope.lookup("b"), [True, False])
    check_jos(scope.lookup("c"), [])
    check_jos(scope.lookup("d"), [], floom=["test", "test2"])


def test_list_atribute_type_violation_1(snippetcompiler):
    snippetcompiler.setup_for_snippet(
        """
entity Jos:
  bool[] bar = true
end
implement Jos using std::none
c = Jos()
"""
    )
    with pytest.raises(ParserException):
        compiler.do_compile()


def test_list_atribute_type_violation_2(snippetcompiler):
    snippetcompiler.setup_for_snippet(
        """
entity Jos:
  bool[] bar = ["x"]
end
implement Jos using std::none
c = Jos()
"""
    )
    with pytest.raises(RuntimeException):
        compiler.do_compile()


def test_list_atribute_type_violation_3(snippetcompiler):
    snippetcompiler.setup_for_snippet(
        """
entity Jos:
  bool[] bar
end
implement Jos using std::none
c = Jos(bar = ["X"])
"""
    )
    with pytest.raises(RuntimeException):
        compiler.do_compile()


def test_issue_235_empty_lists(snippetcompiler):
    snippetcompiler.setup_for_snippet(
        """
entity Test1:

end
implement Test1 using std::none

entity Test2:
end
implement Test2 using std::none

Test1 tests [0:] -- [0:] Test2 tests

t1 = Test1(tests=[])
std::print(t1.tests)
"""
    )
    (_, root) = compiler.do_compile()
    scope = root.get_child("__config__").scope

    assert scope.lookup("t1").get_value().get_attribute("tests").get_value() == []


def test_608_list_to_list(snippetcompiler):
    snippetcompiler.setup_for_snippet(
        """
implementation none for std::Entity:
end

entity A:
    string name
end

entity B:
    string name
end

B.a [1:] -- A

entity C:
    string name
end

C.a [0:] -- A

implement A using none
implement B using none
implement C using none

a1 = A(name="a1")
a2 = A(name="a2")

b1 = B(name="b1")

c1 = C(name="c1")

b1.a = a1
b1.a = c1.a
"""
    )
    (_, scopes) = compiler.do_compile()


def test_608_list_to_single(snippetcompiler):
    snippetcompiler.setup_for_snippet(
        """
implementation none for std::Entity:
end

entity A:
    string name
end

entity B:
    string name
end

B.a [1] -- A

entity C:
    string name
end

C.a [0:] -- A

implement A using none
implement B using none
implement C using none

a1 = A(name="a1")
a2 = A(name="a2")

b1 = B(name="b1")

c1 = C(name="c1")

b1.a = c1.a
"""
    )
    with pytest.raises(AttributeException):
        (_, scopes) = compiler.do_compile()


def test_608_opt_to_list(snippetcompiler):
    snippetcompiler.setup_for_snippet(
        """
implementation none for std::Entity:
end

entity A:
    string name
end

entity B:
    string name
end

B.a [1:] -- A

entity C:
    string name
end

C.a [0:1] -- A

implement A using none
implement B using none
implement C using none

a1 = A(name="a1")
a2 = A(name="a2")

b1 = B(name="b1")

c1 = C(name="c1")

b1.a = a1
b1.a = c1.a
"""
    )
    with pytest.raises(OptionalValueException):
        (_, scopes) = compiler.do_compile()


def test_608_opt_to_single(snippetcompiler):
    snippetcompiler.setup_for_snippet(
        """
implementation none for std::Entity:
end

entity A:
    string name
end

entity B:
    string name
end

B.a [1] -- A

entity C:
    string name
end

C.a [0:1] -- A

implement A using none
implement B using none
implement C using none

a1 = A(name="a1")

b1 = B(name="b1")

c1 = C(name="c1")

b1.a = a1
b1.a = c1.a
"""
    )
    with pytest.raises(OptionalValueException):
        (_, scopes) = compiler.do_compile()


def test_608_opt_to_single_2(snippetcompiler):
    snippetcompiler.setup_for_snippet(
        """
implementation none for std::Entity:
end

entity A:
    string name
end

entity B:
    string name
end

B.a [1] -- A

entity C:
    string name
end

C.a [0:1] -- A

implement A using none
implement B using none
implement C using none

a1 = A(name="a1")

b1 = B(name="b1")

c1 = C(name="c1")

b1.a = a1
b1.a = c1.a

c1.a = a1
"""
    )
    (_, scopes) = compiler.do_compile()


def test_633_default_on_list(snippetcompiler):
    snippetcompiler.setup_for_snippet(
        """
entity Foo:
   list first=[]
   list second=["a", "b"]
   string[] third=["a", "b"]
end

implementation none for std::Entity:
end

implement Foo using none

foo = Foo()
"""
    )
    (_, scopes) = compiler.do_compile()

    root = scopes.get_child("__config__")
    foo = root.lookup("foo").get_value()

    ab = foo.get_attribute("first").get_value()
    assert ab == []

    second = foo.get_attribute("second").get_value()
    assert second == ["a", "b"]

    third = foo.get_attribute("third").get_value()
    assert third == ["a", "b"]


def test_673_in_list(snippetcompiler):
    snippetcompiler.setup_for_snippet(
        """
entity Test:
    string[] attributes
end

implementation test for Test:

end

implement Test using test when "foo" in self.attributes

Test(attributes=["blah", "foo"])
"""
    )
    compiler.do_compile()


def test_552_string_rendering_for_lists(snippetcompiler):
    snippetcompiler.setup_for_snippet(
        """
entity Network:
    string[] tags=[]
end

implement Network using std::none

net1 = Network(tags=["vlan"])
a="Net has tags {{ net1.tags }}"
"""
    )

    (_, scopes) = compiler.do_compile()

    root = scopes.get_child("__config__")
    a = root.lookup("a").get_value()

    assert a == """Net has tags ['vlan']"""


def test_emptylists(snippetcompiler):
    snippetcompiler.setup_for_snippet(
        """
    implement std::Entity using std::none

    a=std::Entity()
    b=std::Entity()
    c=std::Entity()

    a.provides = b.provides
    b.provides = c.provides
    """
    )
    compiler.do_compile()


def test_653_list_attribute_unset(snippetcompiler):
    snippetcompiler.setup_for_error(
        """
        entity Test:
            string[] bla
        end

        Test()

        implement Test using std::none
        """,
        "The object __config__::Test (instantiated at {dir}/main.cf:6) is not complete:"
        " attribute bla ({dir}/main.cf:3) requires 1 values but only 0 are set",
    )



"""
    Copyright 2017 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

# Yacc example

import ply.yacc as yacc
from ply.yacc import YaccProduction

# Get the token map from the lexer. This is required.
from inmanta.parser.plyInmantaLex import tokens, reserved
from inmanta.ast.statements import Literal, Statement
from inmanta.ast import Location, LocatableString, Range, Namespace
from inmanta.ast.statements.generator import For, Constructor
from inmanta.ast.statements.define import DefineEntity, DefineAttribute, DefineImplement, DefineImplementation, DefineRelation, \
    DefineTypeConstraint, DefineTypeDefault, DefineIndex, DefineImport, DefineImplementInherits
from inmanta.ast.constraint.expression import Operator, Not, IsDefined
from inmanta.ast.statements.call import FunctionCall
from inmanta.ast.statements.assign import CreateList, IndexLookup, StringFormat, CreateDict, ShortIndexLookup, MapLookup
from inmanta.ast.variables import Reference, AttributeReference
from inmanta.parser import plyInmantaLex, ParserException
from inmanta.ast.blocks import BasicBlock
import re
import logging
from inmanta.execute.util import NoneValue
from typing import List, Optional


LOGGER = logging.getLogger()


file = "NOFILE"
namespace = None

precedence = (
    ('left', 'OR'),
    ('left', 'AND'),
    ('right', 'NOT'),
    ('right', 'MLS'),
    ('right', 'MLS_END')
)


def attach_lnr(p: YaccProduction, token: int = 1) -> None:
    v = p[0]
    v.location = Location(file, p.lineno(token))
    v.namespace = namespace
    v.lexpos = p.lexpos(token)


def merge_lnr_to_string(p: YaccProduction, starttoken: int = 1, endtoken: int = 2) -> None:
    v = p[0]

    et = p[endtoken]
    endline = et.elnr
    endchar = et.end

    st = p[starttoken]
    if isinstance(st, LocatableString):
        startline = st.lnr
        startchar = st.start
    else:
        startline = et.lnr
        startchar = et.start

    p[0] = LocatableString(v, Range(file, startline, startchar, endline, endchar), endchar, namespace)


def attach_from_string(p: YaccProduction, token: int = 1) -> None:
    v = p[0]
    v.location = p[token].location
    v.namespace = p[token].namespace


def make_none(p: YaccProduction, token: int) -> Literal:
    none = Literal(NoneValue())
    none.location = Location(file, p.lineno(token))
    none.namespace = namespace
    none.lexpos = p.lexpos(token)
    return none


def p_main_collect(p: YaccProduction) -> None:
    """main : top_stmt main"""
    v = p[2]
    v.insert(0, p[1])
    p[0] = v


def p_main_term(p: YaccProduction) -> None:
    'main : top_stmt'
    p[0] = [p[1]]


def p_top_stmt(p: YaccProduction) -> None:
    '''top_stmt : mls
                | entity_def
                | implement_def
                | implementation_def
                | relation
                | statement
                | typedef
                | index
                | import '''
    p[0] = p[1]


#######################
# IMPORT
#######################

def p_import(p: YaccProduction) -> None:
    '''import : IMPORT ns_ref'''
    p[0] = DefineImport(str(p[2]), str(p[2]))
    attach_lnr(p, 1)


def p_import_1(p: YaccProduction) -> None:
    '''import : IMPORT ns_ref AS ID'''
    p[0] = DefineImport(str(p[2]), p[4])
    attach_lnr(p, 1)
#######################
# STMTS
#######################


def p_stmt(p: YaccProduction) -> None:
    '''statement : assign
                | constructor
                | function_call
                | for'''
    p[0] = p[1]


# def p_stmt_err(p):
#     '''statement : list_def
#               | map_def
#               | var_ref
#               | index_lookup'''
#     raise ParserException(file, p[1].location.lnr, lexer.lexpos, "",
#                           msg="expressions are not valid statements, assign this value to a variable to fix this error")


def p_stmt_list_collect(p: YaccProduction) -> None:
    """stmt_list : statement stmt_list"""
    v = p[2]
    v.append(p[1])
    p[0] = v


def p_stmt_list_term(p: YaccProduction) -> None:
    'stmt_list : statement'
    p[0] = [p[1]]


def p_assign(p: YaccProduction) -> None:
    "assign : var_ref '=' operand"
    p[0] = p[1].as_assign(p[3])
    attach_lnr(p, 2)


def p_assign_extend(p: YaccProduction) -> None:
    "assign : var_ref PEQ operand"
    p[0] = p[1].as_assign(p[3], list_only=True)
    attach_lnr(p, 2)


def p_for(p: YaccProduction) -> None:
    "for : FOR ID IN operand ':' block"
    p[0] = For(p[4], p[2], BasicBlock(namespace, p[6]))
    attach_lnr(p, 1)
#######################
# DEFINITIONS
#######################


def p_entity(p: YaccProduction) -> None:
    "entity_def : ENTITY CID ':' entity_body_outer "
    p[0] = DefineEntity(namespace, p[2], p[4][0], [], p[4][1])
    attach_lnr(p)


def p_entity_err_1(p: YaccProduction) -> None:
    "entity_def : ENTITY ID ':' entity_body_outer "
    raise ParserException(p[2].location, str(p[2]), "Invalid identifier: Entity names must start with a capital")


def p_entity_extends(p: YaccProduction) -> None:
    "entity_def : ENTITY CID EXTENDS class_ref_list ':' entity_body_outer "
    p[0] = DefineEntity(namespace, p[2], p[6][0], p[4], p[6][1])
    attach_lnr(p)


def p_entity_extends_err(p: YaccProduction) -> None:
    "entity_def : ENTITY ID EXTENDS class_ref_list ':' entity_body_outer "
    raise ParserException(p[2].location, str(p[2]), "Invalid identifier: Entity names must start with a capital")


def p_entity_body_outer(p: YaccProduction) -> None:
    '''entity_body_outer : mls entity_body END'''
    p[0] = (p[1], p[2])


def p_entity_body_outer_1(p: YaccProduction) -> None:
    '''entity_body_outer : entity_body END '''
    p[0] = (None, p[1])


def p_entity_body_outer_none(p: YaccProduction) -> None:
    '''entity_body_outer : END '''
    p[0] = (None, [])


def p_entity_body_outer_4(p: YaccProduction) -> None:
    '''entity_body_outer : mls END'''
    p[0] = (p[1], [])


def p_entity_body_collect(p: YaccProduction) -> None:
    ''' entity_body : entity_body attr'''
    p[1].append(p[2])
    p[0] = p[1]


def p_entity_body(p: YaccProduction) -> None:
    ''' entity_body : attr'''
    p[0] = [p[1]]


def p_attribute_type(p: YaccProduction) -> None:
    '''attr_type : ns_ref'''
    p[0] = (p[1], False)


def p_attribute_type_opt(p: YaccProduction) -> None:
    "attr_type : ns_ref '?'"
    p[0] = (p[1], True)


def p_attr(p: YaccProduction) -> None:
    "attr : attr_type ID"
    (attr, nullable) = p[1]
    p[0] = DefineAttribute(attr, p[2], None, nullable=nullable)
    attach_lnr(p, 2)


def p_attr_cte(p: YaccProduction) -> None:
    """attr : attr_type ID '=' constant
           | attr_type ID '=' constant_list"""
    (attr, nullable) = p[1]
    p[0] = DefineAttribute(attr, p[2], p[4], nullable=nullable)
    attach_lnr(p, 2)


def p_attr_undef(p: YaccProduction) -> None:
    "attr : attr_type ID '=' UNDEF"
    (attr, nullable) = p[1]
    p[0] = DefineAttribute(attr, p[2], None, remove_default=True, nullable=nullable)
    attach_lnr(p, 2)


def p_attribute_type_multi(p: YaccProduction) -> None:
    "attr_type_multi : ns_ref '[' ']'"
    p[0] = (p[1], False, Location(file, p.lineno(1)))


def p_attribute_type_multi_opt(p: YaccProduction) -> None:
    "attr_type_multi : ns_ref '[' ']' '?'"
    p[0] = (p[1], True, Location(file, p.lineno(1)))


def p_attr_list(p: YaccProduction) -> None:
    "attr : attr_type_multi ID"
    (attr, nullable, location) = p[1]
    p[0] = DefineAttribute(attr, p[2], None, True, nullable=nullable)
    attach_lnr(p, 2)


def p_attr_list_cte(p: YaccProduction) -> None:
    "attr : attr_type_multi ID '=' constant_list"
    (attr, nullable, _) = p[1]
    p[0] = DefineAttribute(attr, p[2], p[4], True, nullable=nullable)
    attach_lnr(p, 3)


def p_attr_list_undef(p: YaccProduction) -> None:
    "attr : attr_type_multi ID '=' UNDEF"
    (attr, nullable, _) = p[1]
    p[0] = DefineAttribute(attr, p[2], None, True, remove_default=True, nullable=nullable)
    attach_lnr(p, 3)


def p_attr_list_null(p: YaccProduction) -> None:
    "attr : attr_type_multi ID '=' NULL"
    (attr, nullable, _) = p[1]
    p[0] = DefineAttribute(attr, p[2], make_none(p, 3), True, nullable=nullable)
    attach_lnr(p, 3)


def p_attr_dict(p: YaccProduction) -> None:
    "attr : DICT ID"
    p[0] = DefineAttribute(p[1], p[2], None)
    attach_lnr(p, 1)


def p_attr_list_dict(p: YaccProduction) -> None:
    "attr : DICT ID '=' map_def"
    p[0] = DefineAttribute(p[1], p[2], p[4])
    attach_lnr(p, 1)


def p_attr_list_dict_null_err(p: YaccProduction) -> None:
    "attr : DICT ID '=' NULL"
    raise ParserException(p[2].location, str(p[2]), "null can not be assigned to dict, did you mean \"dict? %s = null\"" % p[2])


def p_attr_dict_nullable(p: YaccProduction) -> None:
    "attr : DICT '?' ID"
    p[0] = DefineAttribute(p[1], p[3], None, nullable=True)
    attach_lnr(p, 1)


def p_attr_list_dict_nullable(p: YaccProduction) -> None:
    "attr : DICT '?'  ID '=' map_def"
    p[0] = DefineAttribute(p[1], p[3], p[5], nullable=True)
    attach_lnr(p, 1)


def p_attr_list_dict_null(p: YaccProduction) -> None:
    "attr : DICT '?'  ID '=' NULL"
    p[0] = DefineAttribute(p[1], p[3],  make_none(p, 5), nullable=True)
    attach_lnr(p, 1)


# IMPLEMENT
def p_implement_inh(p: YaccProduction) -> None:
    "implement_def : IMPLEMENT class_ref USING PARENTS"
    p[0] = DefineImplementInherits(p[2])
    attach_lnr(p)


def p_implement(p: YaccProduction) -> None:
    "implement_def : IMPLEMENT class_ref USING ns_list"
    p[0] = DefineImplement(p[2], p[4], Literal(True))
    attach_lnr(p)


def p_implement_when(p: YaccProduction) -> None:
    "implement_def : IMPLEMENT class_ref USING ns_list WHEN condition"
    p[0] = DefineImplement(p[2], p[4], p[6])
    attach_lnr(p)


def p_implement_comment(p: YaccProduction) -> None:
    "implement_def : IMPLEMENT class_ref USING ns_list mls"
    p[0] = DefineImplement(p[2], p[4], Literal(True), comment=p[5])
    attach_lnr(p)


def p_implement_inh_comment(p: YaccProduction) -> None:
    "implement_def : IMPLEMENT class_ref USING PARENTS mls"
    p[0] = DefineImplementInherits(p[2], comment=p[5])
    attach_lnr(p)


def p_implement_when_comment(p: YaccProduction) -> None:
    "implement_def : IMPLEMENT class_ref USING ns_list WHEN condition mls"
    p[0] = DefineImplement(p[2], p[4], p[6], comment=p[7])
    attach_lnr(p)


# IMPLEMENTATION


def p_implementation_def(p: YaccProduction) -> None:
    "implementation_def : IMPLEMENTATION ID FOR class_ref implementation"
    docstr, stmts = p[5]
    p[0] = DefineImplementation(namespace, p[2], p[4], BasicBlock(namespace, stmts), docstr)
    attach_lnr(p)


# def p_implementation_def_2(p):
#     "implementation_def : IMPLEMENTATION ID implementation"
#     p[0] = DefineImplementation(namespace, p[2], None, BasicBlock(namespace, p[3]))
#     attach_lnr(p)

def p_implementation(p: YaccProduction) -> None:
    "implementation : ':' mls block"
    p[0] = (p[2], p[3])


def p_implementation_1(p: YaccProduction) -> None:
    "implementation : ':' block"
    p[0] = (None, p[2])


def p_block(p: YaccProduction) -> None:
    "block : stmt_list END"
    p[0] = p[1]


def p_block_empty(p: YaccProduction) -> None:
    "block : END"
    p[0] = []

# RELATION


def p_relation(p: YaccProduction) -> None:
    "relation : class_ref ID multi REL multi class_ref ID"
    if not(p[4] == '--'):
        LOGGER.warning("DEPRECATION: use of %s in relation definition is deprecated, use -- (in %s)" %
                       (p[4], Location(file, p.lineno(4))))
    p[0] = DefineRelation((p[1], p[2], p[3]), (p[6], p[7], p[5]))
    attach_lnr(p, 2)


def p_relation_comment(p: YaccProduction) -> None:
    "relation : class_ref ID multi REL multi class_ref ID mls"
    if not(p[4] == '--'):
        LOGGER.warning("DEPRECATION: use of %s in relation definition is deprecated, use -- (in %s)" %
                       (p[4], Location(file, p.lineno(4))))
    rel = DefineRelation((p[1], p[2], p[3]), (p[6], p[7], p[5]))
    rel.comment = str(p[8])
    p[0] = rel
    attach_lnr(p, 2)


def p_relation_new_outer_comment(p: YaccProduction) -> None:
    "relation : relationnew mls"
    rel = p[1]
    rel.comment = str(p[2])
    p[0] = rel


def p_relation_new_outer(p: YaccProduction) -> None:
    "relation : relationnew"
    p[0] = p[1]


def p_relation_new(p: YaccProduction) -> None:
    "relationnew : class_ref '.' ID multi REL class_ref '.' ID multi"
    p[0] = DefineRelation((p[1], p[8], p[9]), (p[6], p[3], p[4]))
    attach_lnr(p, 2)


def p_relation_new_unidir(p: YaccProduction) -> None:
    "relationnew : class_ref '.' ID multi REL class_ref"
    p[0] = DefineRelation((p[1], None, None), (p[6], p[3], p[4]))
    attach_lnr(p, 2)


def p_relation_new_annotated(p: YaccProduction) -> None:
    "relationnew : class_ref '.' ID multi operand_list class_ref '.' ID multi"
    p[0] = DefineRelation((p[1], p[8], p[9]), (p[6], p[3], p[4]), p[5])
    attach_lnr(p, 2)


def p_relation_new_annotated_unidir(p: YaccProduction) -> None:
    "relationnew : class_ref '.' ID multi operand_list class_ref"
    p[0] = DefineRelation((p[1], None, None), (p[6], p[3], p[4]), p[5])
    attach_lnr(p, 2)


def p_multi_1(p: YaccProduction) -> None:
    "multi : '[' INT ']' "
    p[0] = (p[2], p[2])


def p_multi_2(p: YaccProduction) -> None:
    "multi : '[' INT ':' ']' "
    p[0] = (p[2], None)


def p_multi_3(p: YaccProduction) -> None:
    "multi : '[' INT ':' INT ']' "
    p[0] = (p[2], p[4])


def p_multi_4(p: YaccProduction) -> None:
    "multi : '['  ':' INT ']' "
    p[0] = (None, p[3])

# typedef


def p_typedef_outer(p: YaccProduction) -> None:
    """typedef : typedef_inner"""
    p[0] = p[1]


def p_typedef_outer_comment(p: YaccProduction) -> None:
    """typedef : typedef_inner mls"""
    tdef = p[1]
    tdef.comment = str(p[2])
    p[0] = tdef


def p_typedef_1(p: YaccProduction) -> None:
    """typedef_inner : TYPEDEF ID AS ns_ref MATCHING REGEX
                | TYPEDEF ID AS ns_ref MATCHING condition"""
    p[0] = DefineTypeConstraint(namespace, p[2], p[4], p[6])
    attach_lnr(p, 2)


def p_typedef_cls(p: YaccProduction) -> None:
    """typedef_inner : TYPEDEF CID AS constructor"""
    p[0] = DefineTypeDefault(namespace, p[2], p[4])
    attach_lnr(p, 2)
# index


def p_index(p: YaccProduction) -> None:
    """index : INDEX class_ref '(' id_list ')' """
    p[0] = DefineIndex(p[2], p[4])
    attach_lnr(p, 1)

#######################
# CONDITIONALS


def p_condition_1(p: YaccProduction) -> None:
    "condition : '(' condition ')'"
    p[0] = p[2]


def p_condition_2(p: YaccProduction) -> None:
    """condition : operand CMP_OP operand
                | operand IN list_def
                | operand IN var_ref
                | condition AND condition
                | condition OR condition """
    operator = Operator.get_operator_class(str(p[2]))
    p[0] = operator(p[1], p[3])
    attach_lnr(p, 2)


def p_condition_3(p: YaccProduction) -> None:
    """condition : function_call
                | var_ref"""
    p[0] = p[1]


def p_condition_not(p: YaccProduction) -> None:
    """condition : NOT condition"""
    p[0] = Not(p[2])
    attach_lnr(p)


def p_condition_is_defined(p: YaccProduction) -> None:
    """condition : var_ref '.' ID IS DEFINED"""
    p[0] = IsDefined(p[1], p[3])
    attach_lnr(p, 2)


def p_condition_is_defined_short(p: YaccProduction) -> None:
    """condition : ID IS DEFINED"""
    ref = Reference('self')
    ref.location = p[1].get_location()
    p[0] = IsDefined(ref, p[1])
    attach_lnr(p)


def p_condition_term_1(p: YaccProduction) -> None:
    """condition : TRUE
                | FALSE"""
    p[0] = Literal(p[1])
    attach_lnr(p)

#######################
# EXPRESSIONS


# TODO
def p_operand(p: YaccProduction) -> None:
    """ operand : constant
              | function_call
              | constructor
              | list_def
              | map_def
              | var_ref
              | index_lookup
              | map_lookup"""
    p[0] = p[1]


def p_map_lookup(p: YaccProduction) -> None:
    """ map_lookup : attr_ref '[' operand ']'
                   | local_var '[' operand ']'
                   | map_lookup '[' operand ']'"""
    p[0] = MapLookup(p[1], p[3])


def p_constructor(p: YaccProduction) -> None:
    " constructor : class_ref '(' param_list ')' "
    p[0] = Constructor(p[1], p[3], Location(file, p.lineno(2)), namespace)


def p_function_call(p: YaccProduction) -> None:
    " function_call : ns_ref '(' operand_list ')'"
    p[0] = FunctionCall(str(p[1]), p[3])
    attach_lnr(p, 2)


def p_list_def(p: YaccProduction) -> None:
    " list_def : '[' operand_list ']'"
    p[0] = CreateList(p[2])
    attach_lnr(p, 1)


def p_pair_list_collect(p: YaccProduction) -> None:
    """pair_list : STRING ':' operand ',' pair_list"""
    p[5].insert(0, (str(p[1]), p[3]))
    p[0] = p[5]


def p_pair_list_term(p: YaccProduction) -> None:
    "pair_list : STRING ':' operand"
    p[0] = [(str(p[1]), p[3])]


def p_pair_list_term_2(p: YaccProduction) -> None:
    "pair_list : "
    p[0] = []


def p_map_def(p: YaccProduction) -> None:
    " map_def : '{' pair_list '}'"
    p[0] = CreateDict(p[2])
    attach_lnr(p, 1)


def p_map_def_empty(p: YaccProduction) -> None:
    " map_def : '{' '}'"
    p[0] = CreateDict([])
    attach_lnr(p, 1)


def p_index_lookup(p: YaccProduction) -> None:
    " index_lookup : class_ref '[' param_list ']'"
    p[0] = IndexLookup(p[1], p[3])
    attach_lnr(p, 2)


def p_short_index_lookup(p: YaccProduction) -> None:
    " index_lookup : attr_ref '[' param_list ']'"
    attref = p[1]
    p[0] = ShortIndexLookup(attref.instance, attref.attribute, p[3])
    attach_lnr(p, 2)
#######################
# HELPERS


def p_constant_mls(p: YaccProduction) -> None:
    """ constant : mls """
    p[0] = Literal(str(p[1]))
    attach_from_string(p, 1)


def p_constant(p: YaccProduction) -> None:
    """ constant : INT
    | FLOAT
    """
    p[0] = Literal(p[1])
    attach_lnr(p)


def p_constant_none(p: YaccProduction) -> None:
    """ constant : NULL
    """
    p[0] = make_none(p, 1)
    attach_lnr(p)


def p_constant_regex(p: YaccProduction) -> None:
    """ constant : REGEX
    """
    p[0] = p[1]
    attach_lnr(p)


def p_constant_t(p: YaccProduction) -> None:
    """ constant : TRUE
    """
    p[0] = Literal(True)
    attach_lnr(p)


def p_constant_f(p: YaccProduction) -> None:
    """ constant : FALSE
    """
    p[0] = Literal(False)
    attach_lnr(p)


formatRegex = r"""({{\s*([\.A-Za-z0-9_-]+)\s*}})"""
format_regex_compiled = re.compile(formatRegex, re.MULTILINE | re.DOTALL)


def p_string(p: YaccProduction) -> None:
    " constant : STRING "
    value = p[1]
    match_obj = format_regex_compiled.findall(str(value))

    if len(match_obj) > 0:
        p[0] = create_string_format(value, match_obj, Location(file, p.lineno(1)))
    else:
        p[0] = Literal(str(value))
    attach_lnr(p)


def create_string_format(format_string: str, variables: List[str], location: Location) -> StringFormat:
    """
        Create a string interpolation statement
    """
    _vars = []

    for var_str in variables:
        var_parts = var_str[1].split(".")
        ref = Reference(var_parts[0])
        ref.namespace = namespace

        if len(var_parts) > 1:
            for attr in var_parts[1:]:
                ref = AttributeReference(ref, attr)
                ref.location = location
                ref.namespace = namespace
            _vars.append((ref, var_str[0]))
        else:
            _vars.append((ref, var_str[0]))

    return StringFormat(str(format_string), _vars)


def p_constant_list(p: YaccProduction) -> None:
    " constant_list : '[' constants ']' "
    p[0] = CreateList(p[2])
    attach_lnr(p, 1)


def p_constants_term(p: YaccProduction) -> None:
    "constants : constant"
    p[0] = [p[1]]


def p_constants_term_2(p: YaccProduction) -> None:
    "constants : "
    p[0] = []


def p_constants_collect(p: YaccProduction) -> None:
    """constants : constant ',' constants"""
    p[3].insert(0, p[1])
    p[0] = p[3]


def p_param_list_collect(p: YaccProduction) -> None:
    """param_list : ID '=' operand ',' param_list"""
    p[5].insert(0, (p[1], p[3]))
    p[0] = p[5]


def p_param_list_term(p: YaccProduction) -> None:
    "param_list : ID '=' operand"
    p[0] = [(p[1], p[3])]


def p_param_list_term_2(p: YaccProduction) -> None:
    "param_list : "
    p[0] = []


def p_operand_list_collect(p: YaccProduction) -> None:
    """operand_list : operand ',' operand_list"""
    p[3].insert(0, p[1])
    p[0] = p[3]


def p_operand_list_term(p: YaccProduction) -> None:
    'operand_list : operand'
    p[0] = [p[1]]


def p_operand_list_term_2(p: YaccProduction) -> None:
    "operand_list :"
    p[0] = []


def p_ns_list_collect(p: YaccProduction) -> None:
    """ns_list : ns_ref ',' ns_list"""
    p[3].insert(0, p[1])
    p[0] = p[3]


def p_ns_list_term(p: YaccProduction) -> None:
    'ns_list : ns_ref'
    p[0] = [p[1]]


def p_var_ref(p: YaccProduction) -> None:
    "var_ref : attr_ref"
    p[0] = p[1]


def p_attr_ref(p: YaccProduction) -> None:
    "attr_ref : var_ref '.' ID"
    p[0] = AttributeReference(p[1], p[3])
    attach_lnr(p, 2)


def p_local_var(p: YaccProduction) -> None:
    "local_var : ns_ref"
    p[0] = Reference(p[1])
    attach_from_string(p, 1)


def p_var_ref_2(p: YaccProduction) -> None:
    "var_ref : ns_ref"
    p[0] = Reference(p[1])
    attach_from_string(p, 1)


def p_class_ref_direct(p: YaccProduction) -> None:
    "class_ref : CID"
    p[0] = p[1]


# def p_class_ref_direct_err(p):
#     "class_ref : ID"
#     raise ParserException(
#         file, p.lineno(1), p.lexpos(1), p[1], "Invalid identifier: Entity names must start with a capital")


def p_class_ref(p: YaccProduction) -> None:
    "class_ref : ns_ref SEP CID"
    p[0] = "%s::%s" % (str(p[1]), p[3])
    merge_lnr_to_string(p, 1, 3)


# def p_class_ref_err(p):
#     "class_ref : ns_ref SEP ID"
#     raise ParserException(
#         file, p.lineno(3), p.lexpos(3), p[3], "Invalid identifier: Entity names must start with a capital")


def p_class_ref_list_collect(p: YaccProduction) -> None:
    """class_ref_list : class_ref ',' class_ref_list"""
    p[3].insert(0, p[1])
    p[0] = p[3]


def p_class_ref_list_collect_err(p: YaccProduction) -> None:
    """class_ref_list : var_ref ',' class_ref_list"""
    raise ParserException(p[1].location, str(p[1]), "Invalid identifier: Entity names must start with a capital")


def p_class_ref_list_term(p: YaccProduction) -> None:
    'class_ref_list : class_ref'
    p[0] = [p[1]]


def p_class_ref_list_term_err(p: YaccProduction) -> None:
    'class_ref_list : var_ref'

    raise ParserException(p[1].location, str(p[1]), "Invalid identifier: Entity names must start with a capital")


def p_ns_ref(p: YaccProduction) -> None:
    "ns_ref : ns_ref SEP ID"
    p[0] = "%s::%s" % (p[1], p[3])
    merge_lnr_to_string(p, 1, 3)


def p_ns_ref_term(p: YaccProduction) -> None:
    "ns_ref : ID"
    p[0] = p[1]


def p_id_list_collect(p: YaccProduction) -> None:
    """id_list : ID "," id_list"""
    p[3].insert(0, p[1])
    p[0] = p[3]


def p_id_list_term(p: YaccProduction) -> None:
    'id_list : ID'
    p[0] = [p[1]]


def p_mls_term(p: YaccProduction) -> None:
    "mls : MLS_END"
    p[0] = p[1]


def p_mls_collect(p: YaccProduction) -> None:
    "mls : MLS mls"
    p[0] = "%s%s" % (p[1], p[2])
    merge_lnr_to_string(p, 1, 2)


# Error rule for syntax errors
def p_error(p: YaccProduction) -> None:
    pos = lexer.lexpos - lexer.linestart + 1
    r = Range(file, lexer.lineno, pos, lexer.lineno, pos)

    if p is None:
        # at end of file
        raise ParserException(r, "Unexpected end of file")

    # keyword instead of ID
    if p.type in reserved.values():
        if hasattr(p.value, "location"):
            r = p.value.location
        raise ParserException(r, str(p.value), "invalid identifier, %s is a reserved keyword" % p.value)

    if parser.symstack[-1].type in reserved.values():
        if hasattr(parser.symstack[-1].value, "location"):
            r = parser.symstack[-1].value.location
        raise ParserException(r, str(parser.symstack[-1].value),
                              "invalid identifier, %s is a reserved keyword" % parser.symstack[-1].value)

    raise ParserException(r, p.value)


# Build the parser
lexer = plyInmantaLex.lexer
parser = yacc.yacc()


def myparse(ns: Namespace, tfile: str, content: Optional[str]) -> List[Statement]:
    global file
    file = tfile
    lexer.inmfile = tfile
    global namespace
    namespace = ns
    lexer.namespace = ns
    lexer.begin('INITIAL')

    if content is None:
        with open(tfile, 'r') as myfile:
            data = myfile.read()
            if len(data) == 0:
                return []
            # prevent problems with EOF
            data = data + "\n"
            lexer.lineno = 1
            lexer.linestart = 0
            return parser.parse(data, lexer=lexer, debug=False)
    else:
        data = content
        if len(data) == 0:
            return []
        # prevent problems with EOF
        data = data + "\n"
        lexer.lineno = 1
        lexer.linestart = 0
        return parser.parse(data, lexer=lexer, debug=False)


def parse(namespace: Namespace, filename: str, content: Optional[str] = None) -> List[Statement]:
    statements = myparse(namespace, filename, content)
    # self.cache(filename, statements)
    return statements



"""
    Copyright 2017 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

from inmanta.config import *
import logging

LOGGER = logging.getLogger(__name__)

# flake8: noqa: H904

agent_map = \
    Option("config", "agent-map", None,
           """By default the agent assumes that all agent names map to the host on which the process is executed. With the
agent map it can be mapped to other hosts. This value consists of a list of key/value pairs. The key is the name of the
agent and the format of the value is described in :inmanta:entity:`std::AgentConfig`

example: iaas_openstack=localhost,vm1=192.16.13.2""", is_map)

environment = \
    Option("config", "environment", None,
           "The environment this agent or compile belongs to", is_uuid_opt)

agent_names = \
    Option("config", "agent-names", "$node-name",
           "Names of the agents this instance should deploy configuration for", is_str)

agent_interval = \
    Option("config", "agent-interval", 600, """[DEPRECATED] The run interval of the agent.
Every run-interval seconds, the agent will check the current state of its resources against to desired state model""", is_time)

agent_splay = \
    Option("config", "agent-splay", 600, """[DEPRECATED] The splaytime added to the runinterval.
Set this to 0 to disable splaytime.
 At startup the agent will choose a random number between 0 and "agent_splay.
It will wait this number of second before performing the first deploy.
Each subsequent deploy will start agent-interval seconds after the previous one.""", is_time)

agent_reconnect_delay = \
    Option("config", "agent-reconnect-delay", 5,
           "Time to wait after a failed heartbeat message. DO NOT SET TO 0 ", is_int)

server_timeout = \
    Option("config", "server-timeout", 125,
           "Amount of time to wait for a response from the server before we try to reconnect, must be smaller than server.agent-hold", is_time)

agent_deploy_interval = \
    Option("config", "agent-deploy-interval", 0,
           "The number of seconds between two (incremental) deployment runs of the agent.", is_time, predecessor_option=agent_interval)
agent_deploy_splay_time = \
    Option("config", "agent-deploy-splay-time", 600,
           """The splaytime added to the agent-deploy-interval. Set this to 0 to disable the splaytime.

At startup the agent will choose a random number between 0 and agent-deploy-splay-time.
It will wait this number of second before performing the first deployment run.
Each subsequent repair deployment will start agent-deploy-interval seconds after the previous one.""", is_time,
           predecessor_option=agent_splay)

agent_repair_interval = \
    Option("config", "agent-repair-interval", 600,
           "The number of seconds between two repair runs (full deploy) of the agent. " +
           "Set this to 0 to disable the scheduled repair runs.", is_time)
agent_repair_splay_time = \
    Option("config", "agent-repair-splay-time", 600,
           """The splaytime added to the agent-repair-interval. Set this to 0 to disable the splaytime.

At startup the agent will choose a random number between 0 and agent-repair-splay-time.
It will wait this number of second before performing the first repair run.
Each subsequent repair deployment will start agent-repair-interval seconds after the previous one.""", is_time)



##############################
# agent_rest_transport
##############################

agent_transport = TransportConfig("agent")



"""
    Copyright 2018 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""
import random
import base64
import threading
import os
import time

import pytest
from tornado.httpclient import HTTPRequest, AsyncHTTPClient
from inmanta import config, protocol
from inmanta.protocol.rest import CallArguments
from inmanta.util import hash_file
from inmanta.server import config as opt
from tornado import gen, web
import tornado


def make_random_file(size=0):
    """
        Generate a random file.

        :param size: If size is > 0 content is generated that is equal or more than size.
    """
    randomvalue = str(random.randint(0, 10000))
    if size > 0:
        while len(randomvalue) < size:
            randomvalue += randomvalue

    content = ("Hello world %s\n" % (randomvalue)).encode()
    hash = hash_file(content)

    body = base64.b64encode(content).decode("ascii")

    return (hash, content, body)


@pytest.mark.asyncio
async def test_client_files(client):
    (hash, content, body) = make_random_file()

    # Check if the file exists
    result = await client.stat_file(id=hash)
    assert result.code == 404

    # Create the file
    result = await client.upload_file(id=hash, content=body)
    assert result.code == 200

    # Get the file
    result = await client.get_file(id=hash)
    assert result.code == 200
    assert "content" in result.result
    assert result.result["content"] == body


@pytest.mark.asyncio
async def test_client_files_lost(client):
    (hash, content, body) = make_random_file()

    # Get the file
    result = await client.get_file(id=hash)
    assert result.code == 404


@pytest.mark.asyncio
async def test_sync_client_files(client):
    done = []
    limit = 100
    sleep = 0.01

    def do_test():
        sync_client = protocol.SyncClient("client")

        (hash, content, body) = make_random_file()

        # Check if the file exists
        result = sync_client.stat_file(id=hash)
        assert result.code == 404

        # Create the file
        result = sync_client.upload_file(id=hash, content=body)
        assert result.code == 200

        # Get the file
        result = sync_client.get_file(id=hash)
        assert result.code == 200
        assert "content" in result.result
        assert result.result["content"] == body

        done.append(True)

    thread = threading.Thread(target=do_test)
    thread.start()

    while len(done) == 0 and limit > 0:
        await gen.sleep(sleep)
        limit -= 1

    thread.join()
    assert len(done) > 0


@pytest.mark.asyncio
async def test_client_files_stat(client):

    file_names = []
    i = 0
    while i < 10:
        (hash, content, body) = make_random_file()
        if hash not in file_names:
            file_names.append(hash)
            result = await client.upload_file(id=hash, content=body)
            assert result.code == 200
            i += 1

    result = await client.stat_files(files=file_names)
    assert len(result.result["files"]) == 0

    other_files = ["testtest"]
    result = await client.stat_files(files=file_names + other_files)
    assert len(result.result["files"]) == len(other_files)


@pytest.mark.asyncio
async def test_diff(client):
    ca = "Hello world\n".encode()
    ha = hash_file(ca)
    result = await client.upload_file(id=ha, content=base64.b64encode(ca).decode("ascii"))
    assert result.code == 200

    cb = "Bye bye world\n".encode()
    hb = hash_file(cb)
    result = await client.upload_file(id=hb, content=base64.b64encode(cb).decode("ascii"))
    assert result.code == 200

    diff = await client.diff(ha, hb)
    assert diff.code == 200
    assert len(diff.result["diff"]) == 5

    diff = await client.diff(0, hb)
    assert diff.code == 200
    assert len(diff.result["diff"]) == 4

    diff = await client.diff(ha, 0)
    assert diff.code == 200
    assert len(diff.result["diff"]) == 4


@pytest.mark.asyncio
async def test_client_files_bad(client):
    (hash, content, body) = make_random_file()
    # Create the file
    result = await client.upload_file(id=hash + "a", content=body)
    assert result.code == 400


@pytest.mark.asyncio
async def test_client_files_corrupt(client):
    (hash, content, body) = make_random_file()
    # Create the file
    result = await client.upload_file(id=hash, content=body)
    assert result.code == 200

    state_dir = opt.state_dir.get()

    file_dir = os.path.join(state_dir, "server", "files")

    file_name = os.path.join(file_dir, hash)

    with open(file_name, "wb+") as fd:
        fd.write("Haha!".encode())

    opt.server_delete_currupt_files.set("false")
    result = await client.get_file(id=hash)
    assert result.code == 500

    result = await client.upload_file(id=hash, content=body)
    assert result.code == 500

    opt.server_delete_currupt_files.set("true")
    result = await client.get_file(id=hash)
    assert result.code == 500

    result = await client.upload_file(id=hash, content=body)
    assert result.code == 200


@pytest.mark.asyncio
async def test_gzip_encoding(server):
    """
        Test if the server accepts gzipped encoding and returns gzipped encoding.
    """
    (hash, content, body) = make_random_file(size=1024)

    port = config.Config.get("server_rest_transport", "port")
    url = "http://localhost:%s/api/v1/file/%s" % (port, hash)

    zipped, body = protocol.gzipped_json({"content": body})
    assert zipped

    request = HTTPRequest(
        url=url,
        method="PUT",
        headers={"Accept-Encoding": "gzip", "Content-Encoding": "gzip"},
        body=body,
        decompress_response=True,
    )
    client = AsyncHTTPClient()
    response = await client.fetch(request)
    assert response.code == 200

    request = HTTPRequest(url=url, method="GET", headers={"Accept-Encoding": "gzip"}, decompress_response=True)
    client = AsyncHTTPClient()
    response = await client.fetch(request)
    assert response.code == 200
    assert response.headers["X-Consumed-Content-Encoding"] == "gzip"


class MainHandler(web.RequestHandler):

    def get(self):
        time.sleep(1.1)


@pytest.fixture(scope="function")
async def app(unused_tcp_port):
    http_app = web.Application([(r"/api/v1/file/abc", MainHandler)])
    server = tornado.httpserver.HTTPServer(http_app)
    server.bind(unused_tcp_port)
    server.start()
    yield server

    server.stop()
    await server.close_all_connections()


@pytest.mark.asyncio(timeout=30)
async def test_timeout_error(app):
    """
        Test test verifies that the protocol client can handle requests that timeout. This means it receives a http error
        status that is not generated by the server but by the client.
    """
    from inmanta.config import Config

    Config.load_config()

    port = str(list(app._sockets.values())[0].getsockname()[1])
    Config.set("client_rest_transport", "port", port)
    Config.set("client_rest_transport", "request_timeout", "1")

    from inmanta import protocol

    client = protocol.Client("client")
    x = await client.get_file(id="abc")

    assert x.code == 599
    assert "message" in x.result


@pytest.mark.asyncio
async def test_method_properties():
    """
        Test method properties decorator and helper functions
    """
    @protocol.method(method_name="test", operation="PUT", client_types=["api"])
    def test_method(name):
        """
            Create a new project
        """

    props: protocol.common.MethodProperties = test_method.__method_properties__
    assert "Authorization" in props.get_call_headers()
    assert props.get_listen_url() == "/api/v1/test"
    assert props.get_call_url({}) == "/api/v1/test"


@pytest.mark.asyncio
async def test_invalid_client_type():
    """
        Test invalid client ype
    """
    with pytest.raises(Exception) as e:
        @protocol.method(method_name="test", operation="PUT", client_types=["invalid"])
        def test_method(name):
            """
                Create a new project
            """
        assert "Invalid client type invalid specified for function" in str(e)


@pytest.mark.asyncio
async def test_call_arguments_defaults():
    """
        Test processing RPC messages
    """
    @protocol.method(method_name="test", operation="PUT", client_types=["api"])
    def test_method(name: str, value: int = 10):
        """
            Create a new project
        """

    call = CallArguments(test_method.__method_properties__, {"name": "test"}, {})
    await call.process()

    assert call.call_args["name"] == "test"
    assert call.call_args["value"] == 10


def test_create_client():
    with pytest.raises(AssertionError):
        protocol.SyncClient("agent", "120")

    with pytest.raises(AssertionError):
        protocol.Client("agent", "120")
